"""
isicarchive.numfunc

This module provides numeric helper functions and doesn't have to be
imported from outside the main package functionality (IsicApi).

Functions
---------
calcbetas
    Perform least-squares regression and return regression weights
transform
"""

# specific version for file
__version__ = '0.4.8'


# imports (needed for majority of functions)
from typing import Any, List, Optional, Tuple, Union
import warnings

import numpy


def calcbetas():
    """
    Perform least-squares regression and return regression weights

    Parameters
    ----------
    
    """
    pass
